/*
 * Licensed materials, Property of IBM Corp.
 *
 * openCryptoki CCA token
 *
 * (C) COPYRIGHT International Business Machines Corp. 2001, 2002, 2006
 *
 * Author: Kent E. Yoder <yoder1@us.ibm.com>
 *
 */


// // to get the variant dependency out
#ifndef  _TOK_LOCAL_
#define  _TOK_LOCAL_
#define PK_LITE_DIR  token_specific.token_directory

#define PK_DIR      PK_LITE_DIR
#define SUB_DIR     token_specific.token_subdir


#define DBGTAG  token_specific.token_debug_tag

#endif
